import asyncio
import logging
from typing import Callable, Optional

from .connection_details import VncConnectionDetails
from .display_activity_monitor import start_activity_monitor, stop_activity_monitor
from .utils import run_command

logger = logging.getLogger(__name__)


class PtWebVncCommands:
    @staticmethod
    def start(
        display_id: int,
        window_title: Optional[str] = None,
        ssl_certificate: Optional[str] = None,
        height: Optional[int] = None,
        width: Optional[int] = None,
        depth: Optional[int] = None,
        run: Optional[str] = None,
        background_colour: Optional[str] = None,
        with_window_manager: Optional[bool] = None,
    ) -> str:
        cmd = f"pt-web-vnc start --display-id {display_id} "

        if window_title:
            cmd += f"--window-title {window_title} "
        if ssl_certificate:
            cmd += f"--ssl-certificate {ssl_certificate} "
        if height:
            cmd += f"--height {height} "
        if width:
            cmd += f"--width {width} "
        if depth:
            cmd += f"--depth {depth} "
        if run:
            cmd += f"--run-command {run} "
        if background_colour:
            cmd += f"--background-colour {background_colour} "
        if with_window_manager:
            cmd += "--with-window-manager "
        return cmd

    @staticmethod
    def stop(display_id) -> str:
        return f"pt-web-vnc stop --display-id {display_id}"

    @staticmethod
    def url(display_id) -> str:
        return f"pt-web-vnc url --display-id {display_id}"


def start(
    display_id: int,
    window_title: Optional[str] = None,
    ssl_certificate: Optional[str] = None,
    height: Optional[int] = None,
    width: Optional[int] = None,
    depth: Optional[int] = None,
    run: Optional[str] = None,
    background_colour: Optional[str] = None,
    with_window_manager: Optional[bool] = None,
    on_display_activity: Optional[Callable] = None,
) -> None:
    cmd = PtWebVncCommands.start(
        display_id=display_id,
        window_title=window_title,
        ssl_certificate=ssl_certificate,
        height=height,
        width=width,
        depth=depth,
        run=run,
        background_colour=background_colour,
        with_window_manager=with_window_manager,
    )

    logging.info(f"Starting pt-web-vnc: {cmd}")
    run_command(cmd, timeout=10)

    if callable(on_display_activity):
        raise NotImplementedError(
            "To run the 'on activity' feature, use the 'async_start' function"
        )


def stop(display_id: int) -> None:
    cmd = PtWebVncCommands.stop(display_id)
    logging.info(f"Stopping pt-web-vnc: {cmd}")
    run_command(cmd, timeout=10)


def connection_details(display_id: int) -> VncConnectionDetails:
    cmd = PtWebVncCommands.url(display_id)
    logging.info(f"Getting pt-web-vnc connection details: {cmd}")
    url = run_command(cmd, timeout=10)
    return VncConnectionDetails(url=url.strip())


async def async_start(
    display_id: int,
    window_title: Optional[str] = None,
    ssl_certificate: Optional[str] = None,
    height: Optional[int] = None,
    width: Optional[int] = None,
    depth: Optional[int] = None,
    run: Optional[str] = None,
    background_colour: Optional[str] = None,
    with_window_manager: Optional[bool] = None,
    on_display_activity: Optional[Callable] = None,
) -> None:
    cmd = PtWebVncCommands.start(
        display_id=display_id,
        window_title=window_title,
        ssl_certificate=ssl_certificate,
        height=height,
        width=width,
        depth=depth,
        run=run,
        background_colour=background_colour,
        with_window_manager=True
        if callable(on_display_activity)
        else with_window_manager,
    )

    logging.info(f"Starting pt-web-vnc: {cmd}")
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await proc.wait()
    if callable(on_display_activity):
        start_activity_monitor(
            display_id, on_display_activity, connection_details(display_id)
        )


async def async_stop(display_id: int) -> None:
    cmd = PtWebVncCommands.stop(display_id)
    logging.info(f"Stopping pt-web-vnc: {cmd}")
    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await proc.wait()
    await stop_activity_monitor(display_id)


async def async_connection_details(display_id: int) -> VncConnectionDetails:
    cmd = PtWebVncCommands.url(display_id)
    logging.info(f"Getting pt-web-vnc connection details: {cmd}")

    proc = await asyncio.create_subprocess_shell(
        cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    novnc_url, _ = await proc.communicate()
    return VncConnectionDetails(url=novnc_url.decode().strip())



import asyncio
import logging
from typing import Callable, List, Optional, Union

from .connection_details import VncConnectionDetails


class DisplayActivityMonitor:
    display_number: int
    callbacks: List
    current_activity: Union[None, str]
    on_display_activity: Optional[Callable]
    _stop: bool
    _monitor_task: Optional[asyncio.Task]
    SLEEP_TIME = 1

    def __init__(self, display_number: int) -> None:
        self.display_number = display_number
        self.callbacks = []
        self.current_activity = None
        self.on_display_activity = None
        self._stop = False
        self._monitor_task = None

    def start(self) -> None:
        self._monitor_task = asyncio.create_task(self._monitor())

    async def _monitor(self) -> None:
        logging.debug(
            f"Starting to monitor for activity on display '{self.display_number}'"
        )
        self._stop = False
        self.current_activity = await self._get_number_of_windows()

        while not self._stop:
            current_windows = await self._get_number_of_windows()
            if current_windows and current_windows != self.current_activity:
                logging.debug(f"Activity detected on display '{self.display_number}'")
                self.current_activity = current_windows
                if callable(self.on_display_activity):
                    await self.on_display_activity()
            await asyncio.sleep(self.SLEEP_TIME)

    async def _get_number_of_windows(self):
        proc = await asyncio.create_subprocess_shell(
            f"xwininfo -d :{self.display_number} -root -tree",
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, _ = await proc.communicate()
        return stdout

    async def stop(self) -> None:
        logging.debug(
            f"Stopping checks for activity on display '{self.display_number}'"
        )
        self._stop = True
        if self._monitor_task:
            self._monitor_task.cancel()
            await asyncio.wait(self._monitor_task)


display_activity_monitors = {}


def start_activity_monitor(
    display_id: int,
    on_display_activity: Callable,
    connection_details: VncConnectionDetails,
):
    activity_monitor = DisplayActivityMonitor(display_id)

    # Callback to run when there's activity in a display
    async def on_monitor_activity():
        if connection_details:
            await on_display_activity(connection_details)

    display_activity_monitors[display_id] = activity_monitor
    activity_monitor.on_display_activity = on_monitor_activity
    activity_monitor.start()


async def stop_activity_monitor(display_id: int):
    activity_monitor = display_activity_monitors.get(display_id)
    if activity_monitor:
        await activity_monitor.stop()
        display_activity_monitors.pop(display_id)



from urllib.parse import urlparse


class VncConnectionDetails:
    def __init__(self, url) -> None:
        self._parsed_url = urlparse(url)

    @property
    def url(self):
        return self._parsed_url.geturl()

    @property
    def hostname(self):
        return self._parsed_url.hostname

    @property
    def port(self):
        return self._parsed_url.port

    @property
    def path(self):
        return f"{self._parsed_url.path}?{self._parsed_url.query}"
